--[[
credits:

factorio devs

MagmaMcFry for Factorissimo, which illustrated how to get players, items and power across surfaces

--]]

require('skyscrapers')

--to debug tables; to be removed later
local inspect = require('inspect')

function requestChunkCreationForSkyscraperIndex(index, numberOfFloors)
    for i = 0,numberOfFloors do
        global.skyscrapersSurface.request_to_generate_chunks({index*global.SPACING_SIZE_X + 16, -i*global.SPACING_SIZE_Y-16}, 0)
    end
    global.skyscrapersSurface.force_generate_chunk_requests()
    --maybe it is not needed to block here; it could be done lazily. not sure if worth it.
end

local function init_globals()
    
    global.SPACING_SIZE_X = getSpacingX()
    global.SPACING_SIZE_Y = getSpacingY()

    global.SKYSCRAPERS = getSkyscraperDefinitions()
    global.HIGHESTSKYSCRAPER = getHighestSkyscraper()
    
    --actually it would be possible to allow the creation of them in any (non skyscraper) surface; however most surfaces as of now represent areas where it wouldn't really make sense (e.g. inside a factorissimo)
    --therefore instead use a whitelist that could be extended later
    global.ALLOW_SKYSCRAPERS_ON_SURFACES = {nauvis = true}
    
    global.DOOR_OPENING_SPEED = 10
        
    global.highestSkyscraperIndex = -1
    global.skyscrapersToUpdate = {} --key: surface as string, value: dictionary with key=position.x, value: dictionary with key=position.y, value: skyscraper index, starting with 0
    global.skyscraperIndexToEntityKey = {} --key: skyscraper index, value: {position=entity.position, surface=entity.surface.name, name=entity.name}
    global.elevatorButtons = {}--key: player index, value: top level gui element that contains all others
    global.elevatorRequests = {}--key: skyscraper index, value: {floorNumber=which floor to go to, entity=elevator entity, delayStart=num ticks when to start moving, delayStop=num ticks till open}
    global.elevators = {}--key: skyscraper index, value: elevator entity
    global.tileAnimations = {} --key: surface as string, value: table with key=position.x, value: table with key=position.y, value=name of animation
    global.lastKnownCharacterNames = {} --key: player index, value: the name of the character that was used last
    
    global.skyscrapersSurface = game.create_surface("skyscrapers", {
        --terrain_segmentation
        water="none",
        autoplace_controls = {},
        default_enable_all_autoplace_controls = false,
        autoplace_settings = {
            ["entity"] = { treat_missing_as_default=false, settings={}},
            ["tile"] = { treat_missing_as_default=false, settings={}},
            ["decorative"] = { treat_missing_as_default=false, settings={}},
        },
        --https://lua-api.factorio.com/latest/Concepts.html#CliffPlacementSettings
        cliff_settings={name="cliff", cliff_elevation_interval=40, cliff_elevation_0=1024}, --what about small-cliff though? cannot specify anything for it
        seed = 0,
        --shift 
        width = 0, 
        height = 0, 
        starting_area = "none",
        peaceful_mode = true
    })
                
    global.version = "0.1.1"
                    
end

local function addTileAnimation(surface,x,y,animation)
    if not global.tileAnimations[surface] then global.tileAnimations[surface] = {} end
    if not global.tileAnimations[surface][x] then global.tileAnimations[surface][x] = {} end
    
    local animationDetails = nil
    if animation == "elevator-open" then
        --starts at skyscraper-elevator-entrance-1 and ends at skyscraper-elevator-entrance-4
        animationDetails = {{name="skyscraper-elevator-entrance-2",ticks=4}, {name="skyscraper-elevator-entrance-3", ticks=7}, {name="skyscraper-elevator-entrance-4", ticks=9, 
        audio="elevator_ding"}}
    end
    if animation == "elevator-close" then
        --starts at skyscraper-elevator-entrance-4 and ends at skyscraper-elevator-entrance-1
        animationDetails = {{name="skyscraper-elevator-entrance-3",ticks=2}, {name="skyscraper-elevator-entrance-2", ticks=5}, {name="skyscraper-elevator-entrance-1", ticks=8}}
    end
    if not animationDetails then return end
    
    global.tileAnimations[surface][x][y] = animationDetails
end

script.on_init(function()
    init_globals()
end)

commands.add_command("create_skyscraper_template", "creates a template for the skyscraper that the current player is standing in, and writes it to the script output directory.", function(param)
    cmd_create_skyscraper_template(param)
end)

local function updateModVersion()
    --nothing to update yet
end

script.on_configuration_changed(function()
    updateModVersion()
end)

local function getAbsolutePositionOnFloor(skyscraperIndex, floorIndex, relativePosition)
    return {
        x=skyscraperIndex * global.SPACING_SIZE_X + relativePosition.x,
        y=-floorIndex * global.SPACING_SIZE_Y - relativePosition.y
    }
end

local function enter_skyscraper(player, skyscraper)
    local skyScraperIndex = global.skyscrapersToUpdate[skyscraper.surface.name][skyscraper.position.x][skyscraper.position.y]
    local skyScraperTemplate = global.SKYSCRAPERS[skyscraper.name]
    
    local centerPos = {skyScraperIndex * global.SPACING_SIZE_X + skyScraperTemplate.exitPos.x, -skyScraperTemplate.exitPos.y - 2.0}
    local nonCollidingPos = global.skyscrapersSurface.find_non_colliding_position(player.character.prototype.name, centerPos, 2, 0.1)
    local teleportPos = nonCollidingPos
    if not teleportPos then teleportPos = centerPos end
    player.teleport(teleportPos,global.skyscrapersSurface)
end

local function leave_skyscraper(player)
    local skyScraperIndex = math.floor(player.position.x / global.SPACING_SIZE_X)
    local skyscraperKey = global.skyscraperIndexToEntityKey[skyScraperIndex]
    local skyScraperTemplate = global.SKYSCRAPERS[skyscraperKey.name]
    
    local centerPos = {skyscraperKey.position.x + skyScraperTemplate.entrancePos.x, skyscraperKey.position.y + skyScraperTemplate.entrancePos.y + 2.0}
    local nonCollidingPos = global.skyscrapersSurface.find_non_colliding_position(player.character.prototype.name, centerPos, 2, 0.1)
    local teleportPos = nonCollidingPos
    if not teleportPos then teleportPos = centerPos end
    player.teleport(teleportPos,skyscraperKey.surface)
end

local function skyscraperStairUp(player)
    local skyScraperIndex = math.floor(player.position.x / global.SPACING_SIZE_X)
    local floorIndex = math.floor((-player.position.y) / global.SPACING_SIZE_Y)
    local skyscraperKey = global.skyscraperIndexToEntityKey[skyScraperIndex]
    local skyScraperTemplate = global.SKYSCRAPERS[skyscraperKey.name]
    
    
    local centerPos = {global.SPACING_SIZE_X * skyScraperIndex + skyScraperTemplate.stairsDownPos.x + 1.5,
                -global.SPACING_SIZE_Y * (floorIndex+1) - skyScraperTemplate.stairsDownPos.y + 3.5}
    local nonCollidingPos = global.skyscrapersSurface.find_non_colliding_position(player.character.prototype.name, centerPos, 2, 0.1)
    local teleportPos = nonCollidingPos
    if not teleportPos then teleportPos = centerPos end
    player.teleport(teleportPos)
end

local function skyscraperStairDown(player)
    local skyScraperIndex = math.floor(player.position.x / global.SPACING_SIZE_X)
    local floorIndex = math.floor((-player.position.y) / global.SPACING_SIZE_Y)
    local skyscraperKey = global.skyscraperIndexToEntityKey[skyScraperIndex]
    local skyScraperTemplate = global.SKYSCRAPERS[skyscraperKey.name]
    
    
    local centerPos = {global.SPACING_SIZE_X * skyScraperIndex + skyScraperTemplate.stairsUpPos.x - 0.5,
                -global.SPACING_SIZE_Y * (floorIndex-1) - skyScraperTemplate.stairsUpPos.y + 3.5}
    local nonCollidingPos = global.skyscrapersSurface.find_non_colliding_position(player.character.prototype.name, centerPos, 2, 0.1)
    local teleportPos = nonCollidingPos
    if not teleportPos then teleportPos = centerPos end
    player.teleport(teleportPos)
end

local function find_skyscraper_entrance(surface, playerPosition)
    local candidates = surface.find_entities_filtered{area={{playerPosition.x-0.1,playerPosition.y-0.3},{playerPosition.x+0.1,playerPosition.y}}}
    for _,entity in pairs(candidates) do
        if global.skyscrapersToUpdate[entity.surface.name] and 
           global.skyscrapersToUpdate[entity.surface.name][entity.position.x] and
           global.skyscrapersToUpdate[entity.surface.name][entity.position.x][entity.position.y] ~= nil then
            --check if positioned in the middle where the door is. y is already checked, now test x
            local skyScraperTemplate = global.SKYSCRAPERS[entity.name]
            if playerPosition.x > entity.position.x + skyScraperTemplate.entrancePos.x - 0.25 and playerPosition.x < entity.position.x + skyScraperTemplate.entrancePos.x + 0.25 then return entity end
        end
    end
    return nil
end

local function enterElevator(player)

    local skyScraperIndex = math.floor(player.position.x / global.SPACING_SIZE_X)
    local floorIndex = math.floor((-player.position.y) / global.SPACING_SIZE_Y)
    local skyscraperKey = global.skyscraperIndexToEntityKey[skyScraperIndex]
    local skyScraperTemplate = global.SKYSCRAPERS[skyscraperKey.name]

    local distance=1
    local elevatorEntities = player.surface.find_entities_filtered{name="skyscrapers-elevator", area={{player.position.x-distance,player.position.y-distance},{player.position.x+distance,player.position.y+distance}}}
    local elevator = nil
    for _, entity in pairs(elevatorEntities) do elevator = entity end
    if not elevator then
        if global.elevatorRequests[skyScraperIndex] and global.elevatorRequests[skyScraperIndex].floorNumber == floorIndex then
            --already called here
            return
        end
        if global.elevatorRequests[skyScraperIndex] then
            player.print("waiting for elevator to arrive at its destination before calling it")
            return
        end
        player.print("calling elevator to current floor")
        global.elevatorRequests[skyScraperIndex] = {floorNumber=floorIndex, entity=global.elevators[skyScraperIndex], delayStart=global.DOOR_OPENING_SPEED, delayStop=global.DOOR_OPENING_SPEED, playerIndex=player.index}    
        return 
    end
    
    if global.elevatorRequests[skyScraperIndex] then return end --don't enter an elevator that is already moving

    elevator.get_fuel_inventory().insert({name="coal"}) --refill every time someone enters to ensure it is full
    elevator.set_passenger(player)
    
    if not global.elevatorButtons[player.index] then
        global.elevatorButtons[player.index] = player.gui.top.add{type="frame", direction ="vertical"}
        for i=0,9 do
            local inner = global.elevatorButtons[player.index].add{type="frame", direction ="horizontal"}
            for j=0,9 do
                if (i*10+j>0) and (i*10+j <= skyScraperTemplate.Z) then
                    inner.add{type="button", name=tostring(i*10+j-1), caption=tostring(i)..tostring(j)}
                end
            end
        end
    else
        global.elevatorButtons[player.index].enabled = true
    end
end

local function teleport_players()
    
    local tick = game.tick
    for player_index, player in pairs(game.players) do
        if player.connected and not player.driving then--and tick % 20 == 4 then
            local walking_state = player.walking_state
            if walking_state.walking then
                if walking_state.direction == defines.direction.north
                or walking_state.direction == defines.direction.northeast
                or walking_state.direction == defines.direction.northwest then

                    local skyscraper_entrance = find_skyscraper_entrance(player.surface, player.position)
                    if skyscraper_entrance then
                        enter_skyscraper(player, skyscraper_entrance)
                    end
                    
                    local tile = player.surface.get_tile(player.position.x, player.position.y)
                    if tile.name == "skyscraper-stairs-up" then
                        skyscraperStairUp(player)
                    end
                    if tile.name == "skyscraper-stairs-down" then
                        skyscraperStairDown(player)
                    end
                    if tile.name == "skyscraper-elevator-entrance-1" or  tile.name == "skyscraper-elevator-entrance-2" or  tile.name == "skyscraper-elevator-entrance-3" or  tile.name == "skyscraper-elevator-entrance-4" then
                        enterElevator(player)
                    end
                elseif walking_state.direction == defines.direction.south
                or walking_state.direction == defines.direction.southeast
                or walking_state.direction == defines.direction.southwest then
                    local tile = player.surface.get_tile(player.position.x, player.position.y)
                    if tile.name == "skyscraper-exit" then
                        leave_skyscraper(player)
                    end
                end
                
            end
        end
    end
end

script.on_event(defines.events.on_gui_click, function(event)
    local clicked = event.element
    local playerIndex = event.player_index
    if clicked.parent and clicked.parent.parent and clicked.parent.parent == global.elevatorButtons[playerIndex] then
        local destinationFloor = tonumber(clicked.name)
        global.elevatorButtons[playerIndex].destroy()
        global.elevatorButtons[playerIndex] = nil
        if game.players[playerIndex].vehicle then --possible that the player left via commands, or left manually before pressing a button
            local skyScraperIndex = math.floor(game.players[playerIndex].vehicle.position.x / global.SPACING_SIZE_X)
            global.elevatorRequests[skyScraperIndex] = {floorNumber=destinationFloor, entity=game.players[playerIndex].vehicle, delayStart=global.DOOR_OPENING_SPEED, delayStop=global.DOOR_OPENING_SPEED, playerIndex=playerIndex}       
        end
    end
end)

script.on_event(defines.events.on_gui_closed, function(event)
    if event.entity and event.entity.surface == global.skyscrapersSurface then
        local skyScraperIndex = math.floor(event.entity.position.x / global.SPACING_SIZE_X)
        local skyscraperKey = global.skyscraperIndexToEntityKey[skyScraperIndex]
        local skyScraperTemplate = global.SKYSCRAPERS[skyscraperKey.name]
        if skyScraperTemplate.beacon then
            local beaconTemplateName = "skyscrapers-beacon-template-"..tostring(skyScraperTemplate.beaconModules).."-"..tostring(skyScraperTemplate.beaconEfficiency)
            if event.entity.name == beaconTemplateName then
                local beaconInstanceName = "skyscrapers-beacon-instance-"..tostring(skyScraperTemplate.beaconModules).."-"..tostring(skyScraperTemplate.beaconEfficiency)
                local moduleInv = event.entity.get_module_inventory()
                for i = 0,skyScraperTemplate.Z-1 do
                    local beaconInstances = global.skyscrapersSurface.find_entities_filtered{name=beaconInstanceName,
                        position=getAbsolutePositionOnFloor(global.highestSkyscraperIndex,i,{x=math.ceil(skyScraperTemplate.X/2.0),y=math.ceil(skyScraperTemplate.Y/2.0)})}
                    if #beaconInstances == 1 then
                        local beaconInstance = beaconInstances[1]
                        local moduleInv2 = beaconInstance.get_module_inventory()
                        moduleInv2.clear()
                        for i = 1, #moduleInv do
                            moduleInv2[i].set_stack(moduleInv[i])
                        end
                    end            
                end
            end
        end        
    end
end)

local function updateElevators()
    local keysToRemove = {}
    for requestKey,request in pairs (global.elevatorRequests) do

        local elevator = request.entity
        local skyScraperIndex = math.floor(elevator.position.x / global.SPACING_SIZE_X)
        local floorIndex = math.floor((-elevator.position.y) / global.SPACING_SIZE_Y)
        local skyscraperKey = global.skyscraperIndexToEntityKey[skyScraperIndex]
        local skyScraperTemplate = global.SKYSCRAPERS[skyscraperKey.name]
        local destinationY = -global.SPACING_SIZE_Y * request.floorNumber - skyScraperTemplate.elevatorPos.y
    
        --update combinators with current position
        for i = 0,skyScraperTemplate.Z-1 do
            local combinatorEntities = elevator.surface.find_entities_filtered{name="constant-combinator", position={
            global.SPACING_SIZE_X * skyScraperIndex + skyScraperTemplate.elevatorPos.x + 0.5,
            -global.SPACING_SIZE_Y * i - skyScraperTemplate.elevatorPos.y - 2.0 + 0.5}}
            local combinator = nil
            for _, entity in pairs(combinatorEntities) do combinator = entity end
            if combinator then
                combinator.get_control_behavior().set_signal(1, {
                signal = {type = "virtual", name = "signal-E"},
                count = floorIndex + 1,
                })
            end
        end
    
        if request.delayStart > 0 then 
            if request.delayStart == global.DOOR_OPENING_SPEED then 
                addTileAnimation(elevator.surface.name, elevator.position.x, elevator.position.y, "elevator-close") 
            end 
            request.delayStart = request.delayStart - 1 
            return 
        end
        
        if math.abs(elevator.position.y - destinationY) < 0.1 then
            elevator.teleport({elevator.position.x,destinationY})
            elevator.speed = 0
            if request.delayStop > 0 then
                if request.delayStop == global.DOOR_OPENING_SPEED then 
                    addTileAnimation(elevator.surface.name, elevator.position.x, elevator.position.y, "elevator-open") 
                end
                request.delayStop = request.delayStop - 1 
                return 
            end            
            local player = elevator.get_passenger()
            elevator.set_passenger(nil)
            if player then
                local centerPos = {elevator.position.x,destinationY + 2.0}
                local nonCollidingPos = global.skyscrapersSurface.find_non_colliding_position(global.lastKnownCharacterNames[request.playerIndex], centerPos, 2, 0.1)
                local teleportPos = nonCollidingPos
                if not teleportPos then teleportPos = centerPos end
                player.teleport(teleportPos)
            end
            table.insert(keysToRemove, skyScraperIndex)
        else
            local elevatorSpeed = skyScraperTemplate.elevatorTopSpeed * math.abs(elevator.position.y - destinationY) / 32
            elevatorSpeed = math.min(elevatorSpeed,skyScraperTemplate.elevatorTopSpeed)
            elevatorSpeed = math.max(elevatorSpeed,0.2)
            if elevator.position.y < destinationY then elevator.speed = -elevatorSpeed end
            if elevator.position.y > destinationY then elevator.speed = elevatorSpeed end
        end
        
    end
    for _,v in pairs(keysToRemove) do global.elevatorRequests[v] = nil end
end

local function processTileAnimations()
    local animsToRemove = {}
    for surface,animOnSurface in pairs(global.tileAnimations) do
        for x,animOnX in pairs(animOnSurface) do
            for y, animationDetails in pairs(animOnX) do
                local maxTicks = 0
                for _, tileAnim in pairs(animationDetails) do
                    maxTicks = math.max(maxTicks, tileAnim.ticks)
                    tileAnim.ticks = tileAnim.ticks - 1
                    if tileAnim.ticks == 0 then
                        game.surfaces[surface].set_tiles({{name=tileAnim.name, position={x,y}}}, false)
                        if tileAnim.audio then
                            game.surfaces[surface].play_sound{path=tileAnim.audio, position={x=x,y=y}, volume_modifier=1}
                        end
                    end
                end
                if maxTicks == 0 then
                    table.insert(animsToRemove, {surface=surface,x=x,y=y})
                end
            end
        end
    end
    for _,val in pairs(animsToRemove) do
        global.tileAnimations[val.surface][val.x][val.y] = nil 
    end
end

local function updatePower()
    for skyScraperIndex, v in pairs(global.skyscraperIndexToEntityKey) do
        
        local skyScraperTemplate = global.SKYSCRAPERS[v.name]
        
        local surfaceEntity = game.surfaces[v.surface].find_entity(v.name, v.position)
        local posToSearch = getAbsolutePositionOnFloor(
            skyScraperIndex,0,{x=skyScraperTemplate.accumulatorPos.x + 0.5,y=skyScraperTemplate.accumulatorPos.y - 0.5})
        local skyscraperEntity = global.skyscrapersSurface.find_entity(v.name.."-accumulator", posToSearch)
        
        if surfaceEntity ~= nil and skyscraperEntity ~= nil then
            skyscraperEntity.energy = (surfaceEntity.energy + skyscraperEntity.energy)/2
            surfaceEntity.energy = skyscraperEntity.energy
            --todo maybe something would be needed here to smoothen an animation
        end
    
    end
end

local function updateBeltsNS(entityKey, surfacePos, skyScraperPos)
    local surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="transport-belt",limit=1}
    if #surfaceEntities == 0 then surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="underground-belt",limit=1} 
        if #surfaceEntities > 0 and surfaceEntities[1].belt_to_ground_type == "input" then surfaceEntities = {} end
    end
    local skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="transport-belt",limit=1}
    if #skysSkraperEntities == 0 then skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="underground-belt",limit=1}
        if #skysSkraperEntities > 0 and skysSkraperEntities[1].belt_to_ground_type == "input" then skysSkraperEntities = {} end
    end
    
    if #surfaceEntities > 0 and #skysSkraperEntities > 0 then
        local from = nil
        local to = nil
        if surfaceEntities[1].direction == defines.direction.north and skysSkraperEntities[1].direction == defines.direction.north then
            from = surfaceEntities[1]
            to = skysSkraperEntities[1]
        end
        if surfaceEntities[1].direction == defines.direction.south and skysSkraperEntities[1].direction == defines.direction.south then
            from = skysSkraperEntities[1]
            to = surfaceEntities[1]
        end
        if from and to then
            for beltSide = 1,2 do
                if to.get_transport_line(beltSide).can_insert_at_back() and from.get_transport_line(beltSide).get_item_count() > 0 then
                    local itemToMove = from.get_transport_line(beltSide)[1]
                    to.get_transport_line(beltSide).insert_at_back(itemToMove)
                    from.get_transport_line(beltSide).remove_item(itemToMove)
                end
            end
        end
    end       
end

local function updateBeltsWE(entityKey, surfacePos, skyScraperPos)
    local surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="transport-belt",limit=1}
    if #surfaceEntities == 0 then surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="underground-belt",limit=1} 
        if #surfaceEntities > 0 and surfaceEntities[1].belt_to_ground_type == "input" then surfaceEntities = {} end
    end
    local skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="transport-belt",limit=1}
    if #skysSkraperEntities == 0 then skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="underground-belt",limit=1}
        if #skysSkraperEntities > 0 and skysSkraperEntities[1].belt_to_ground_type == "input" then skysSkraperEntities = {} end
    end
        
    if #surfaceEntities > 0 and #skysSkraperEntities > 0 then
        local from = nil
        local to = nil
        if surfaceEntities[1].direction == defines.direction.east and skysSkraperEntities[1].direction == defines.direction.east then
            from = surfaceEntities[1]
            to = skysSkraperEntities[1]
        end
        if surfaceEntities[1].direction == defines.direction.west and skysSkraperEntities[1].direction == defines.direction.west then
            from = skysSkraperEntities[1]
            to = surfaceEntities[1]
        end
        if from and to then
            for beltSide = 1,2 do
                if to.get_transport_line(beltSide).can_insert_at_back() and from.get_transport_line(beltSide).get_item_count() > 0 then
                    local itemToMove = from.get_transport_line(beltSide)[1]
                    to.get_transport_line(beltSide).insert_at_back(itemToMove)
                    from.get_transport_line(beltSide).remove_item(itemToMove)
                end
            end
        end
    end       
end

local function updateBeltsEW(entityKey, surfacePos, skyScraperPos)
    local surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="transport-belt",limit=1}
    if #surfaceEntities == 0 then surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="underground-belt",limit=1} 
        if #surfaceEntities > 0 and surfaceEntities[1].belt_to_ground_type == "input" then surfaceEntities = {} end
    end
    local skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="transport-belt",limit=1}
    if #skysSkraperEntities == 0 then skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="underground-belt",limit=1}
        if #skysSkraperEntities > 0 and skysSkraperEntities[1].belt_to_ground_type == "input" then skysSkraperEntities = {} end
    end
        
    if #surfaceEntities > 0 and #skysSkraperEntities > 0 then
        local from = nil
        local to = nil
        if surfaceEntities[1].direction == defines.direction.west and skysSkraperEntities[1].direction == defines.direction.west then
            from = surfaceEntities[1]
            to = skysSkraperEntities[1]
        end
        if surfaceEntities[1].direction == defines.direction.east and skysSkraperEntities[1].direction == defines.direction.east then
            from = skysSkraperEntities[1]
            to = surfaceEntities[1]
        end
        if from and to then
            for beltSide = 1,2 do
                if to.get_transport_line(beltSide).can_insert_at_back() and from.get_transport_line(beltSide).get_item_count() > 0 then
                    local itemToMove = from.get_transport_line(beltSide)[1]
                    to.get_transport_line(beltSide).insert_at_back(itemToMove)
                    from.get_transport_line(beltSide).remove_item(itemToMove)
                end
            end
        end
    end       
end

local function updatePipes(entityKey, surfacePos, skyScraperPos)
    local surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="pipe",limit=1}
    if #surfaceEntities == 0 then surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="pipe-to-ground",limit=1} end
    if #surfaceEntities == 0 then surfaceEntities = game.surfaces[entityKey.surface].find_entities_filtered{position=surfacePos,type="storage-tank",limit=1} end
    local skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="pipe",limit=1}
    if #skysSkraperEntities == 0 then skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="pipe-to-ground",limit=1} end
    if #skysSkraperEntities == 0 then skysSkraperEntities = global.skyscrapersSurface.find_entities_filtered{position=skyScraperPos,type="storage-tank",limit=1} end
    
    if #surfaceEntities > 0 and #skysSkraperEntities > 0 then
        local box1 = surfaceEntities[1].fluidbox
        local box2 = skysSkraperEntities[1].fluidbox
        local fluid1 = box1[1]
        local fluid2 = box2[1]
        if not fluid1 and fluid2 then fluid1 = {name=fluid2.name, amount=0, temperature=fluid2.temperature} end
        if fluid1 and not fluid2 then fluid2 = {name=fluid1.name, amount=0, temperature=fluid1.temperature} end
        if fluid1 and fluid2 and fluid1.type == fluid2.type and fluid1.amount + fluid2.amount > 1 then
            local globalRatio = (fluid1.amount + fluid2.amount) / (box1.get_capacity(1) + box2.get_capacity(1))
            local newAmount1 = globalRatio * box1.get_capacity(1)
            local newAmount2 = globalRatio * box2.get_capacity(1)
            local transferredAmount = newAmount1 - fluid1.amount
            if fluid1.temperature and fluid2.temperature then
                local oldTemp1 = fluid1.temperature
                local oldTemp2 = fluid2.temperature
                fluid1.temperature = ((oldTemp1 * fluid1.amount ) + (oldTemp2 * transferredAmount )) / newAmount1
                fluid2.temperature = ((oldTemp2 * fluid2.amount ) - (oldTemp1 * transferredAmount )) / newAmount2
            end
            fluid1.amount = newAmount1
            fluid2.amount = newAmount2
            box1[1] = fluid1
            box2[1] = fluid2
        end
    end       
end

local function updateBeltsAndPipes()

    for skyScraperIndex, v in pairs(global.skyscraperIndexToEntityKey) do
        
        local skyScraperTemplate = global.SKYSCRAPERS[v.name]
        
        local surfaceEntity = game.surfaces[v.surface].find_entity(v.name, v.position)
        if surfaceEntity then
        
            for xOffset = 0, skyScraperTemplate.X - 3 do
                local surfacePos = {x = 1.5 + xOffset + surfaceEntity.position.x - skyScraperTemplate.X / 2, y = 0.5 + surfaceEntity.position.y + skyScraperTemplate.Y / 2}
                local skyScraperPos = {x = xOffset + global.SPACING_SIZE_X * skyScraperIndex + 1.5, y = -0.5}
                
                updateBeltsNS(v, surfacePos, skyScraperPos)
                updatePipes(v, surfacePos, skyScraperPos)
            end
            for yOffset = 0, skyScraperTemplate.Y - 3 do
                local surfacePos = {x = -0.5 + surfaceEntity.position.x - skyScraperTemplate.X / 2, y = -1.5 -yOffset + surfaceEntity.position.y + skyScraperTemplate.Y / 2}
                local skyScraperPos = {x = global.SPACING_SIZE_X * skyScraperIndex + 1.5, y = -0.5 - yOffset}
                
                updateBeltsWE(v, surfacePos, skyScraperPos)
                updatePipes(v, surfacePos, skyScraperPos)
            end
            for yOffset = 0, skyScraperTemplate.Y - 3 do
                local surfacePos = {x = 0.5 + surfaceEntity.position.x + skyScraperTemplate.X / 2, y = -1.5 -yOffset + surfaceEntity.position.y + skyScraperTemplate.Y / 2}
                local skyScraperPos = {x = global.SPACING_SIZE_X * skyScraperIndex - 1.5 + skyScraperTemplate.X, y = -0.5 - yOffset}
                
                updateBeltsEW(v, surfacePos, skyScraperPos)
                updatePipes(v, surfacePos, skyScraperPos)
            end
        end
    end
end

local function updateLastKnownCharacters()
    for _,p in pairs(game.players) do
        if p.character and p.character.prototype and p.character.prototype.name then global.lastKnownCharacterNames[p.index] = p.character.prototype.name end
    end
end

local function spreadPollution()
    
    for index, value in pairs(global.skyscraperIndexToEntityKey) do
    
        local skyScraperTemplate = global.SKYSCRAPERS[value.name]
        local pollutionSum = 0
        for i = 0,skyScraperTemplate.Z - 1 do
            local pos = {index*global.SPACING_SIZE_X + 8, - i * global.SPACING_SIZE_Y - 8}
            local diff = global.skyscrapersSurface.get_pollution(pos)
            global.skyscrapersSurface.pollute(pos, -diff)
            pollutionSum = pollutionSum + diff
        end
        game.surfaces[value.surface].pollute(value.position, pollutionSum)
    end
       
end

script.on_event(defines.events.on_tick, function(event)

    if(event.tick == 0) then global.skyscrapersSurface.request_to_generate_chunks({0,0},4) end

    teleport_players()
    
    updateElevators()
    
    processTileAnimations()
        
    updateBeltsAndPipes()

    updatePower()
    
    updateLastKnownCharacters()
    
    spreadPollution()
            
    
end)

local function abortPlacement(entity, playerBuild, playerIndex, errorMessage)
    local entityPos = {x=entity.position.x, y=entity.position.y}
    local surface = entity.surface
    if playerBuild then
        if entity.name ~= "entity-ghost" then
            game.players[playerIndex].insert{name=entity.name, count=1}
        end
        game.players[playerIndex].print(errorMessage)
    else
        if entity.name ~= "entity-ghost" then
            surface.spill_item_stack(entity.position, {name=entity.name, count=1}, true)
        end
        for onePlayerIndex, player in pairs(game.players) do
            game.players[onePlayerIndex].print(errorMessage)
        end
    end
    entity.destroy()
    for _, ghost in pairs(surface.find_entities_filtered{type="entity-ghost", position=entityPos}) do
        ghost.destroy()
    end
end

local function CheckEntityPlacementInSkyscraper(entity, playerBuild, playerIndex, robot, itemStack, tick)
    --always accept basic entites
    if entity.type == "transport-belt" or entity.type == "underground-belt" or entity.type == "splitter" or entity.type == "pipe" or entity.type == "pipe-to-ground"
        or entity.type == "container" or entity.type == "inserter" or entity.type == "loader" 
        or entity.name == "SNTD-nixie-tube" or entity.name == "SNTD-nixie-tube-small" --allow entities for which build events are explicitly raised
        then
        return
    end
    if entity.name == "entity-ghost" then
        if entity.ghost_type == "transport-belt" or entity.ghost_type == "underground-belt" or entity.ghost_type == "splitter" or entity.ghost_type == "pipe" or entity.ghost_type == "pipe-to-ground"
            or entity.ghost_type == "container" or entity.ghost_type == "inserter" or entity.ghost_type == "loader" 
            or entity.ghost_name == "SNTD-nixie-tube" or entity.ghost_name == "SNTD-nixie-tube-small" --allow entities for which build events are explicitly raised
        then
            return
        end
    end
    
    local skyScraperIndex = math.floor(entity.position.x / global.SPACING_SIZE_X)
    local skyscraperKey = global.skyscraperIndexToEntityKey[skyScraperIndex]
    local skyScraperTemplate = global.SKYSCRAPERS[skyscraperKey.name]
    local allowedCraftingCategories = skyScraperTemplate.allowedCraftingCategories
    if not allowedCraftingCategories then return end
    local actualCraftingCategories = nil
    if entity.name == "entity-ghost" then
        actualCraftingCategories = entity.ghost_prototype.crafting_categories
    else
        actualCraftingCategories = entity.prototype.crafting_categories
    end
    
    local str = ""
    for _,v in pairs(allowedCraftingCategories) do str  = str .. v .. ";" end
    errString = "entity cannot be placed in skyscraper, allowed are only belts(normal,underground,splitter,loader), inserters, pipes, chests, and entities that can craft " .. str
    
    if not actualCraftingCategories then
        abortPlacement(entity, playerBuild, playerIndex, errString )
        return
    end
    
    local foundMatch = false
    --note the intential key/value mismatch: according to documentation: Note: The value in the dictionary is meaningless and exists just to allow the dictionary type for easy lookup.
    --(however allowedCraftingCategories is just a list created by the mod itself, therefore its value is the actual text)
    for v,_ in pairs(actualCraftingCategories) do
        for _,v2 in pairs(allowedCraftingCategories) do
            if v == v2 then foundMatch = true end
        end
    end
    
    if not foundMatch then
        abortPlacement(entity, playerBuild, playerIndex, errString )
        return
    end

end

local function checkCreatedEntity(entity, playerBuild, playerIndex, robot, itemStack, tick)
    if entity.surface == global.skyscrapersSurface then
        CheckEntityPlacementInSkyscraper(entity, playerBuild, playerIndex, robot, itemStack, tick)
        return
    end
    if global.SKYSCRAPERS[entity.name] then
        
        local skyscraperTemplate = global.SKYSCRAPERS[entity.name]
    
        if not ( global.ALLOW_SKYSCRAPERS_ON_SURFACES[entity.surface.name]) then
            abortPlacement(entity, playerBuild, playerIndex, {"item-limitation.not-allowed-on-this-surface"})
            return
        end
        if skyscraperTemplate.unique then
            local entityCount = 0
            for _,surface in pairs(game.surfaces) do
                entityCount = entityCount + surface.count_entities_filtered{name=entity.name}
            end
            if entityCount > 1 then
                abortPlacement(entity, playerBuild, playerIndex, {"item-limitation.only-allowed-once"})
                return
            end
        end
                
        global.highestSkyscraperIndex = global.highestSkyscraperIndex + 1
        requestChunkCreationForSkyscraperIndex(global.highestSkyscraperIndex, skyscraperTemplate.Z)
        if not global.skyscrapersToUpdate[entity.surface.name] then global.skyscrapersToUpdate[entity.surface.name] = {} end
        if not global.skyscrapersToUpdate[entity.surface.name][entity.position.x] then global.skyscrapersToUpdate[entity.surface.name][entity.position.x] = {} end
        global.skyscrapersToUpdate[entity.surface.name][entity.position.x][entity.position.y] = global.highestSkyscraperIndex
        global.skyscraperIndexToEntityKey[global.highestSkyscraperIndex] = {position=entity.position, surface=entity.surface.name, name=entity.name}
        
        local tilesToCreate = {}
        for y = 0,-global.SPACING_SIZE_Y * (global.HIGHESTSKYSCRAPER) +1, -1 do
            for x = global.SPACING_SIZE_X * global.highestSkyscraperIndex,global.SPACING_SIZE_X * (global.highestSkyscraperIndex + 1 )- 1 do                
                table.insert(tilesToCreate, {name="out-of-map", position = {x,y}})
            end
        end 
        global.skyscrapersSurface.set_tiles(tilesToCreate, false)
        
        tilesToCreate = {}
        for i = 0,skyscraperTemplate.Z-1 do
            for y = -global.SPACING_SIZE_Y * i, -global.SPACING_SIZE_Y * i - skyscraperTemplate.Y + 1, -1 do
                for x = global.SPACING_SIZE_X * global.highestSkyscraperIndex, global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.X - 1 do                
                    if x == global.SPACING_SIZE_X * global.highestSkyscraperIndex 
                    or x == global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.X - 1 
                    or y == -global.SPACING_SIZE_Y * i
                    or y ==  -global.SPACING_SIZE_Y * i - skyscraperTemplate.Y + 1 then
                        table.insert(tilesToCreate, {name="skyscraper-boundary", position = {x,y}})
                    else
                        table.insert(tilesToCreate, {name=skyscraperTemplate.tile, position = {x,y}})
                    end
                end
            end
        end
        global.skyscrapersSurface.set_tiles(tilesToCreate, true)
        global.skyscrapersSurface.set_tiles({{name="skyscraper-exit", position={
            global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.exitPos.x,
            -skyscraperTemplate.exitPos.y}}}, false)
            
        tilesToCreate = {}
        for i = 0,skyscraperTemplate.Z-1 do
            if i ~= skyscraperTemplate.Z-1 then
                table.insert(tilesToCreate, {name="skyscraper-stairs-up", position = {global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.stairsUpPos.x,
                -global.SPACING_SIZE_Y * i - skyscraperTemplate.stairsUpPos.y}})
            end
            if i ~= 0 then
                table.insert(tilesToCreate, {name="skyscraper-stairs-down", position = {global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.stairsDownPos.x,
                -global.SPACING_SIZE_Y * i - skyscraperTemplate.stairsDownPos.y}})
            end
        end
        global.skyscrapersSurface.set_tiles(tilesToCreate, true)
        
        for i = 0,skyscraperTemplate.Z-1 do
            local substation = global.skyscrapersSurface.create_entity{name="skyscrapers-substation", 
                position=getAbsolutePositionOnFloor(global.highestSkyscraperIndex,i,{x=math.ceil(skyscraperTemplate.X/2.0),y=math.ceil(skyscraperTemplate.Y/2.0)}), 
                force="neutral"}
            local pole1 = global.skyscrapersSurface.create_entity{name="skyscrapers-pole", 
                position=getAbsolutePositionOnFloor(global.highestSkyscraperIndex,i,{x=math.ceil(skyscraperTemplate.X/2.0) - 1,y=math.ceil(skyscraperTemplate.Y/2.0)}), 
                force="neutral"}
            local pole2 = global.skyscrapersSurface.create_entity{name="skyscrapers-pole", 
                position=getAbsolutePositionOnFloor(global.highestSkyscraperIndex,i,{x=math.ceil(skyscraperTemplate.X/2.0) - 1,y=math.ceil(skyscraperTemplate.Y/2.0) + math.ceil(global.SPACING_SIZE_Y/2.0)}), 
                force="neutral"}
        end
        
        for i = 0,skyscraperTemplate.Z-1 do
            
            local floorTube = global.skyscrapersSurface.create_entity{name="SNTD-nixie-tube", position={
                global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.floorIndicatorPos.x+1,
                -global.SPACING_SIZE_Y * i - skyscraperTemplate.floorIndicatorPos.y}, force="neutral"}
                        
            if playerBuild then
                script.raise_event(defines.events.on_built_entity, { name=defines.events.on_built_entity, created_entity = floorTube, player_index = playerIndex, stack=itemStack, tick=tick})
            else
                script.raise_event(defines.events.on_robot_built_entity, { name=defines.events.on_robot_built_entity, created_entity = floorTube, robot=robot, stack=itemStack, tick=tick})
            end
            
            local floorTube2 = global.skyscrapersSurface.create_entity{name="SNTD-nixie-tube", position={
                global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.floorIndicatorPos.x,
                -global.SPACING_SIZE_Y * i - skyscraperTemplate.floorIndicatorPos.y}, force="neutral"}
                        
            if playerBuild then
                script.raise_event(defines.events.on_built_entity, { name=defines.events.on_built_entity, created_entity = floorTube2, player_index = playerIndex, stack=itemStack, tick=tick})
            else
                script.raise_event(defines.events.on_robot_built_entity, { name=defines.events.on_robot_built_entity, created_entity = floorTube2, robot=robot, stack=itemStack, tick=tick})
            end
                        
            local floorCombinator = global.skyscrapersSurface.create_entity{name="constant-combinator", position={
                global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.floorIndicatorPos.x,
                -global.SPACING_SIZE_Y * i - skyscraperTemplate.floorIndicatorPos.y + 1}, force="neutral"}
                
            floorCombinator.connect_neighbour({wire=defines.wire_type.red,target_entity=floorTube})
                        
            floorCombinator.get_control_behavior().set_signal(1, {
                signal = {type = "virtual", name = "signal-F"},
                count = i+1,
            })
            
            --use > here in order to ensure the 'shortcut' code path does not omit updating this tube! (yes its a hack)
            floorTube.get_control_behavior().circuit_condition = {condition={comparator=">",
                                           first_signal={type="virtual", name="signal-F"},
                                           constant=99}}
   
        end
        
        if skyscraperTemplate.elevator then
        
            local elevatorEntity = global.skyscrapersSurface.create_entity{name="skyscrapers-elevator", position={
                    global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.elevatorPos.x + 0.5,
                    - skyscraperTemplate.elevatorPos.y + 0.5}, force="player"}
            global.elevators[global.highestSkyscraperIndex] = elevatorEntity
            elevatorEntity.get_fuel_inventory().insert({name="coal"})
            --create artificial player to block the driver seat, in order to prevent the player hijacking the elevator and driving around in it
            elevatorEntity.set_driver(global.skyscrapersSurface.create_entity{name = "player", position = elevatorEntity.position, force="player"})
            for i = 0,skyscraperTemplate.Z-1 do
                local tileName = "skyscraper-elevator-entrance-1"
                if i == 0 then tileName = "skyscraper-elevator-entrance-4" end
                global.skyscrapersSurface.set_tiles({{name=tileName, position={
                    global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.elevatorPos.x,
                    -global.SPACING_SIZE_Y * i - skyscraperTemplate.elevatorPos.y}}}, false)
                    
                local elevatorTube = global.skyscrapersSurface.create_entity{name="SNTD-nixie-tube-small", position={
                    global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.elevatorPos.x,
                    -global.SPACING_SIZE_Y * i - skyscraperTemplate.elevatorPos.y - 1.0}, force="neutral"}
                            
                if playerBuild then
                    script.raise_event(defines.events.on_built_entity, { name=defines.events.on_built_entity, created_entity = elevatorTube, player_index = playerIndex, stack=itemStack, tick=tick})
                else
                    script.raise_event(defines.events.on_robot_built_entity, { name=defines.events.on_robot_built_entity, created_entity = elevatorTube, robot=robot, stack=itemStack, tick=tick})
                end
                
                local elevatorCombinator = global.skyscrapersSurface.create_entity{name="constant-combinator", position={
                    global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.elevatorPos.x,
                    -global.SPACING_SIZE_Y * i - skyscraperTemplate.elevatorPos.y - 2.0}, force="neutral"}
                    
                elevatorCombinator.connect_neighbour({wire=defines.wire_type.red,target_entity=elevatorTube})
                            
                elevatorCombinator.get_control_behavior().set_signal(1, {
                    signal = {type = "virtual", name = "signal-E"},
                    count = 1,
                })
                
                --use > here in order to ensure the 'shortcut' code path does not omit updating this tube! (yes its a hack)
                elevatorTube.get_control_behavior().circuit_condition = {condition={comparator=">",
                                               first_signal={type="virtual", name="signal-E"},
                                               constant=99}}
            end
                                           
        end
        for i = 0,skyscraperTemplate.Z-1 do
            local ceilingLight = global.skyscrapersSurface.create_entity{name="small-lamp", position={
                global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.lampPos.x,
                -global.SPACING_SIZE_Y * i - skyscraperTemplate.lampPos.y}, force="neutral"}
        end

        
        local accumulatorEntity = global.skyscrapersSurface.create_entity{name=entity.name .. "-accumulator", position={
                global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.accumulatorPos.x + 0.5,
                - skyscraperTemplate.accumulatorPos.y + 0.5}, force="neutral"}
        accumulatorEntity.energy = skyscraperTemplate.Z * 4000 * 20

        if skyscraperTemplate.beacon then
            local beaconTemplateEntity = global.skyscrapersSurface.create_entity{name="skyscrapers-beacon-template-"..tostring(skyscraperTemplate.beaconModules).."-"..tostring(skyscraperTemplate.beaconEfficiency), position={
                global.SPACING_SIZE_X * global.highestSkyscraperIndex + skyscraperTemplate.beaconTemplatePos.x + 0.5,
                - skyscraperTemplate.beaconTemplatePos.y + 0.5}, force="neutral"}
            for i = 0,skyscraperTemplate.Z-1 do
                local beaconInstanceEntity = global.skyscrapersSurface.create_entity{name="skyscrapers-beacon-instance-"..tostring(skyscraperTemplate.beaconModules).."-"..tostring(skyscraperTemplate.beaconEfficiency),
                    position=getAbsolutePositionOnFloor(global.highestSkyscraperIndex,i,{x=math.ceil(skyscraperTemplate.X/2.0),y=math.ceil(skyscraperTemplate.Y/2.0)}), 
                    force="neutral"}
            end
        end
                
        local prebuildSkyscrapers = getPremadeSkyscraperDefinitions()
        if prebuildSkyscrapers[itemStack.name] then
            for i = 0,skyscraperTemplate.Z-1 do
                local instructions = nil
                if i == 0 then 
                    instructions = prebuildSkyscrapers[itemStack.name].floor1 
                elseif i == skyscraperTemplate.Z-1 then 
                    instructions = prebuildSkyscrapers[itemStack.name].floorTop
                else 
                    instructions = prebuildSkyscrapers[itemStack.name].floorN 
                end
                if instructions then
                    for _,instruction in pairs(instructions) do
                        local positionToCreate = {x = global.SPACING_SIZE_X * global.highestSkyscraperIndex + instruction.x,
                                                  y = -global.SPACING_SIZE_Y * i - instruction.y}
                        local createdEntity = nil
                        if instruction.underground_type then
                            createdEntity = global.skyscrapersSurface.create_entity{name=instruction.name, position=positionToCreate, force="player", direction=instruction.direction, type=instruction.underground_type}
                        elseif instruction.recipe then
                            createdEntity = global.skyscrapersSurface.create_entity{name=instruction.name, position=positionToCreate, force="player", direction=instruction.direction, recipe=instruction.recipe}
                        elseif instruction.inventory_bar then
                            createdEntity = global.skyscrapersSurface.create_entity{name=instruction.name, position=positionToCreate, force="player", bar=instruction.inventory_bar}
                        else
                            createdEntity = global.skyscrapersSurface.create_entity{name=instruction.name, position=positionToCreate, force="player", direction=instruction.direction}
                        end
                        if createdEntity.type == "splitter" then
                            createdEntity.splitter_input_priority = instruction.splitter_input_priority
                            createdEntity.splitter_output_priority = instruction.splitter_output_priority
                        end
                    end
                end
            end
        end
    end
end

script.on_event(defines.events.on_built_entity, function(event)
    checkCreatedEntity(event.created_entity, true, event.player_index, nil, event.stack, event.tick)
end)

script.on_event(defines.events.on_robot_built_entity, function(event)
    checkCreatedEntity(event.created_entity, false, -1, event.robot, event.stack, event.tick)
end)

local function checkMinedEntity(entity)
    if global.skyscrapersToUpdate[entity.surface.name]
        and global.skyscrapersToUpdate[entity.surface.name][entity.position.x]
        and global.skyscrapersToUpdate[entity.surface.name][entity.position.x][entity.position.y] ~= nil then
    
        local skyscraperIndex = global.skyscrapersToUpdate[entity.surface.name][entity.position.x][entity.position.y]
        global.skyscraperIndexToEntityKey[skyscraperIndex] = nil
        global.skyscrapersToUpdate[entity.surface.name][entity.position.x][entity.position.y] = nil
    
        local remainder = entity.surface.create_entity{name="skyscraper-remnants", 
        position=entity.position, 
        force="player"}
        local remainderInv = remainder.get_inventory(defines.inventory.chest)
        
        local stuffInSkyscraper = global.skyscrapersSurface.find_entities({{skyscraperIndex*global.SPACING_SIZE_X, -global.HIGHESTSKYSCRAPER*global.SPACING_SIZE_Y}, {(skyscraperIndex+1)*global.SPACING_SIZE_X - 1, 0}})
        for _, stuff in pairs(stuffInSkyscraper) do
            if stuff.prototype.mineable_properties.products then
                for _,product in pairs(stuff.prototype.mineable_properties.products) do
                    if product.type == "item" 
                    and stuff.name ~= "SNTD-nixie-tube-sprite" 
                    and stuff.name ~= "SNTD-nixie-tube-small-sprite" 
                    --and stuff.name ~= "SNTD-nixie-tube" 
                    --and stuff.name ~= "SNTD-nixie-tube-small" 
                    then
                        local amount = product.amount
                        if amount == nil then
                            if product.probability - math.random() >= 0 then
                                amount = math.random * (product.amount_max - product.amount_min) + product.amount_min
                            end
                        end
                        --insert may fail if it doesn't fit, which is intended.
                        remainderInv.insert({name=product.name, amount=amount})
                    end
                end 
                if stuff.can_be_destroyed() 
                    and stuff.name ~= "SNTD-nixie-tube-sprite" 
                    and stuff.name ~= "SNTD-nixie-tube-small-sprite" 
                    and stuff.name ~= "SNTD-nixie-tube" 
                    and stuff.name ~= "SNTD-nixie-tube-small" then
                    stuff.destroy()
                end
            end
        end
    end
end

script.on_event(defines.events.on_player_mined_entity, function(event)
    checkMinedEntity(event.entity)
end)

script.on_event(defines.events.on_robot_mined_entity, function(event)
    checkMinedEntity(event.entity)
end)

--script.on_event(defines.events.on_entity_died, function(event)
--    checkMinedEntity(event.entity)
--end)

local function preventSpecialTilesFromBeingOverwritten(oldTilesAndPosition, surface)
    for _, oldTileAndPosition in pairs(oldTilesAndPosition) do
        local tileName = oldTileAndPosition.old_tile.name
        if tileName == "skyscraper-exit" 
        or tileName == "skyscraper-stairs-up"
        or tileName == "skyscraper-stairs-down"
        or tileName == "skyscraper-elevator-entrance-1"
        or tileName == "skyscraper-elevator-entrance-2"
        or tileName == "skyscraper-elevator-entrance-3"
        or tileName == "skyscraper-elevator-entrance-4" then
            surface.set_tiles({{name=tileName, position=oldTileAndPosition.position}}, false)
        end
    end
end

script.on_event(defines.events.on_player_built_tile, function(event)
    preventSpecialTilesFromBeingOverwritten(event.tiles, game.surfaces[event.surface_index])
end)

script.on_event(defines.events.on_robot_built_tile, function(event)
    preventSpecialTilesFromBeingOverwritten(event.tiles, event.robot.surface)
end)

function scanForTemplateCreation(boundingBox)
    local inserterEntities = global.skyscrapersSurface.find_entities_filtered{area=boundingBox, type="inserter"}
    local assemblingEntities = global.skyscrapersSurface.find_entities_filtered{area=boundingBox, type="assembling-machine"}
    local furnaceEntities = global.skyscrapersSurface.find_entities_filtered{area=boundingBox, type="furnace"}
    local containerEntities = global.skyscrapersSurface.find_entities_filtered{area=boundingBox, type="container"}
    local beltEntities = global.skyscrapersSurface.find_entities_filtered{area=boundingBox, type="transport-belt"}
    local splitterEntities = global.skyscrapersSurface.find_entities_filtered{area=boundingBox, type="splitter"}
    local undergroundEntities = global.skyscrapersSurface.find_entities_filtered{area=boundingBox, type="underground-belt"}
    local entityList = {}
    for _,e in pairs(inserterEntities) do table.insert(entityList, e) end
    for _,e in pairs(assemblingEntities) do table.insert(entityList, e) end
    for _,e in pairs(furnaceEntities) do table.insert(entityList, e) end
    for _,e in pairs(containerEntities) do table.insert(entityList, e) end
    for _,e in pairs(beltEntities) do table.insert(entityList, e) end
    for _,e in pairs(splitterEntities) do table.insert(entityList, e) end
    for _,e in pairs(undergroundEntities) do table.insert(entityList, e) end
    local str = ""
    local directionToString = {}
    directionToString[defines.direction.north] = "defines.direction.north"
    directionToString[defines.direction.east] = "defines.direction.east"
    directionToString[defines.direction.south] = "defines.direction.south"
    directionToString[defines.direction.west] = "defines.direction.west"
    for _, entity in pairs(entityList) do
        local nameStr = "name=\"" .. entity.name .. "\""
        local xStr = ",x=\"" .. tostring(entity.position.x - boundingBox.left_top[1]) .. "\""
        local yStr = ",y=\"" .. tostring(-(entity.position.y - boundingBox.right_bottom[2])) .. "\""
        local directionStr = ",direction=" .. directionToString[entity.direction] .. ""

        local recipeStr = ""
        if entity.type == "assembling-machine" and entity.get_recipe() then
            recipeStr = ",recipe=\"" .. entity.get_recipe().name .. "\""
        end
        
        local inventoryBarStr = ""
        if entity.type == "container" then
            --condition necessary? even unrestricted chests have a bar value
            inventoryBarStr = ",inventory_bar=\"" .. tostring(entity.get_inventory(defines.inventory.chest).getbar()) .. "\""
        end
        
        local underGroundTypeStr = ""
        if entity.type == "underground-belt" then
            underGroundTypeStr = ",underground_type=\"" .. entity.belt_to_ground_type .. "\""
        end
        
        local splitterPrioStr = ""
        if entity.type == "splitter" then
            splitterPrioStr = ",splitter_input_priority=\"" .. entity.splitter_input_priority .. "\"" .. ",splitter_output_priority=\"" .. entity.splitter_output_priority .. "\""
        end
        
        str = str .. "        {" .. nameStr .. xStr .. yStr .. directionStr .. recipeStr .. inventoryBarStr .. underGroundTypeStr .. splitterPrioStr .. "},\n"
    end
    return str
end

function cmd_create_skyscraper_template(data)
    if data.player_index and game.players[data.player_index].surface == global.skyscrapersSurface then
        local skyScraperIndex = math.floor(game.players[data.player_index].position.x / global.SPACING_SIZE_X)
        local skyScraperKey = global.skyscraperIndexToEntityKey[skyScraperIndex] --key: skyscraper index, value: {position=entity.position, surface=entity.surface.name, name=entity.name}
        local skyScraperTemplate = global.SKYSCRAPERS[skyScraperKey.name]
    
        local floor1Data = scanForTemplateCreation({
            left_top = {skyScraperIndex*global.SPACING_SIZE_X, -skyScraperTemplate.Y}, 
            right_bottom = {skyScraperIndex*global.SPACING_SIZE_X + skyScraperTemplate.X, 0}})
        local floorNData = scanForTemplateCreation({
            left_top = {skyScraperIndex*global.SPACING_SIZE_X, -skyScraperTemplate.Y - global.SPACING_SIZE_Y}, 
            right_bottom = {skyScraperIndex*global.SPACING_SIZE_X + skyScraperTemplate.X, -global.SPACING_SIZE_Y}})
        local floorTopData = scanForTemplateCreation({
            left_top = {skyScraperIndex*global.SPACING_SIZE_X, -skyScraperTemplate.Y - (global.SPACING_SIZE_Y*(skyScraperTemplate.Z-1))}, 
            right_bottom = {skyScraperIndex*global.SPACING_SIZE_X + skyScraperTemplate.X, -(global.SPACING_SIZE_Y*(skyScraperTemplate.Z-1))}})
    
        local data = "return {\n" ..
        "    base=\"" .. skyScraperKey.name .. "\",\n" ..
        "    techToUnlock=\"" .. skyScraperTemplate.techToUnlock .. "\",\n" ..
        "    itemName=\"TODO fill name of item whose icon is rendered on item\",\n" ..
        "    textName=\"TODO fill the name of the skyscraper template\",\n" ..
        "    textDescription=\"TODO fill the description of the skyscraper template\",\n" ..
        "    floor1={\n" ..
        floor1Data ..
        "    },\n" ..
        "    floorN={\n" ..
        floorNData ..
        "    },\n" ..
        "    floorTop={\n" ..
        floorTopData ..
        "    }\n" ..
        "}\n"
        game.write_file("skyscraper_template_data.lua", data, false)
    end    
end
